@@ -27,7 +27,7 @@ class DeserializationError(Exception):
     pass
 
 
-def _object_ref_deserializer(binary, owner_address):
+def _object_ref_deserializer(binary, owner_address, object_status):
     # NOTE(suquark): This function should be a global function so
     # cloudpickle can access it directly. Otherwise cloudpickle
     # has to dump the whole function definition, which is inefficient.
@@ -51,7 +51,7 @@ def _object_ref_deserializer(binary, owner_address):
         if outer_id is None:
             outer_id = ray.ObjectRef.nil()
         worker.core_worker.deserialize_and_register_object_ref(
-            obj_ref.binary(), outer_id, owner_address)
+            obj_ref.binary(), outer_id, owner_address, object_status)
     return obj_ref
 
 
@@ -88,9 +88,10 @@ class SerializationContext:
             self.add_contained_object_ref(obj)
             worker = ray.worker.global_worker
             worker.check_connected()
-            obj, owner_address = (
+            obj, owner_address, object_status = (
                 worker.core_worker.serialize_and_promote_object_ref(obj))
-            return _object_ref_deserializer, (obj.binary(), owner_address)
+            return _object_ref_deserializer, \
+                (obj.binary(), owner_address, object_status)
 
         self._register_cloudpickle_reducer(ray.ObjectRef, object_ref_reducer)
         serialization_addons.apply(self)
@@ -118,11 +119,9 @@ class SerializationContext:
     def set_out_of_band_serialization(self):
         self._thread_local.in_band = False
 
-    def set_outer_object_ref(self, outer_object_ref):
-        self._thread_local.outer_object_ref = outer_object_ref
-
     def get_outer_object_ref(self):
-        return getattr(self._thread_local, "outer_object_ref", None)
+        stack = getattr(self._thread_local, "object_ref_stack", [])
+        return stack[-1] if stack else None
 
     def get_and_clear_contained_object_refs(self):
         if not hasattr(self._thread_local, "object_refs"):
@@ -239,19 +238,25 @@ class SerializationContext:
 
     def deserialize_objects(self, data_metadata_pairs, object_refs):
         assert len(data_metadata_pairs) == len(object_refs)
+        # initialize the thread-local field
+        if not hasattr(self._thread_local, "object_ref_stack"):
+            self._thread_local.object_ref_stack = []
         results = []
         for object_ref, (data, metadata) in zip(object_refs,
                                                 data_metadata_pairs):
-            assert self.get_outer_object_ref() is None
-            self.set_outer_object_ref(object_ref)
             try:
+                # Push the object ref to the stack, so the object under
+                # the object ref knows where it comes from.
+                self._thread_local.object_ref_stack.append(object_ref)
                 obj = self._deserialize_object(data, metadata, object_ref)
             except Exception as e:
                 logger.exception(e)
                 obj = RaySystemError(e, traceback.format_exc())
+            finally:
+                # Must clear ObjectRef to not hold a reference.
+                if self._thread_local.object_ref_stack:
+                    self._thread_local.object_ref_stack.pop()
             results.append(obj)
-            # Must clear ObjectRef to not hold a reference.
-            self.set_outer_object_ref(None)
         return results
 
     def _serialize_to_pickle5(self, metadata, value):
